# Copyright (c) 2014, Nicolas Vanhoren
# 
# Released under the MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals, print_function, absolute_import

import sqlalchemy as sa
import datetime
import re
import asgard
import werkzeug.security

WERKZEUG_METHOD = "pbkdf2:sha512:10000"

def encode_werkzeug(password, method=WERKZEUG_METHOD):
    return "werkzeug:://" + werkzeug.security.generate_password_hash(password, method)

class UsersPlugin(asgard.Plugin):

    config_key = "users"
    dependencies = []

    def __init__(self, app):
        plugin = self
        self.app = app
        self.config = {
            "preferred_encryption" : "werkzeug",
            "bcrypt_turns" : 10,
            "werkzeug_method" : WERKZEUG_METHOD,
        }
        self.users = sa.Table('users', app.metadata,
            sa.Column('id', sa.Integer, primary_key=True),
            sa.Column('email', sa.String(50), nullable=True, unique=True),
            sa.Column('password_hash', sa.String(200), nullable=False),
            sa.Column('creation_date', sa.DateTime(), nullable=False, default=datetime.datetime.now),
        )

        @app.manager
        class UsersManager(asgard.table_manager(self.users)):

            def encode_password(self, password):
                return self._encode_password(password)

            def _encode_password(self, password):
                password = unicode(password)
                if plugin.config["preferred_encryption"] == "bcrypt":
                    import bcrypt
                    password = password.encode("utf8")
                    phash = bcrypt.hashpw(password, bcrypt.gensalt(plugin.config["bcrypt_turns"]))
                    return "bcrypt:://" + phash
                elif plugin.config["preferred_encryption"] == "werkzeug":
                    return encode_werkzeug(password, plugin.config["werkzeug_method"])
                else:
                    raise ValueError("unknown encryption")

            def _check_password(self, password, to_compare):
                password = unicode(password)
                found = re.match("^(\w+)\:\:\/\/(.*)$", to_compare)
                if found.group(1) == "bcrypt":
                    import bcrypt
                    password = password.encode("utf8")
                    phash = unicode(found.group(2)).encode("utf8")
                    if bcrypt.hashpw(password, phash) == phash:
                        return True
                    else:
                        return False
                if found.group(1) == "werkzeug":
                    return werkzeug.security.check_password_hash(found.group(2), password)
                else:
                    raise ValueError("unknown encryption")

            def create_user(self, email, password):
                hash_ = self._encode_password(password)
                return self.create({"email": email, "password_hash": hash_})

            def set_password(self, user_id, password):
                hash_ = self._encode_password(password)
                updated = self.update_by_id(user_id, {"password_hash": hash_})
                return updated

            def test_user(self, email, password):
                try:
                    hash_ = self.read_by_email(email, ["password_hash"])["password_hash"]
                except asgard.PersistenceException:
                    return False
                return self._check_password(password, hash_)

            def read_by_email(self, email, fields=None):
                users = self.read(["email == :email", {"email": email}], fields)
                if len(users) == 0:
                    raise asgard.PersistenceException("Unknown email: %s" % email)
                return users[0]

        self.UsersManager = UsersManager

    def configure(self, config):
        self.config.update(config)



# Copyright (c) 2014, Nicolas Vanhoren
# 
# Released under the MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals, print_function, absolute_import

import werkzeug
import werkzeug.local
import werkzeug.contrib.sessions
import sjoh
import contextlib
import sqlalchemy as sa
import sqlalchemy.sql as sql
import datetime

def create_sessions_table(metadata):
    t = sa.Table('sessions', metadata,
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('session_id', sa.String(50), unique=True, nullable=False),
        sa.Column('data', sa.Text()),
        sa.Column('creation_date', sa.DateTime(), nullable=False, default=datetime.datetime.now),
        sa.Column('last_modification_date', sa.DateTime(), nullable=False, default=datetime.datetime.now),
    )
    return t

class DbSessionStore(werkzeug.contrib.sessions.SessionStore):
    def __init__(self, app):
        super(DbSessionStore, self).__init__()
        self.app = app
        self.serializer = sjoh.JsonSerializer()

    def delete(self, session):
        pass

    def get(self, sid):
        with self.app.transaction():
            if not self.is_valid_key(sid):
                return self.new()
            res = self.app.conn.execute(sql.select([self.app.sessions_table]).where(self.app.sessions_table.c.session_id == sid)).fetchall()
            if len(res) == 0:
                return self.session_class({}, sid, True)
            data = self.serializer.parse(res[0]["data"])
            return self.session_class(data, sid, False)

    def save(self, session):
        """Save a session."""
        with self.app.transaction():
            data = dict(session)
            data = self.serializer.stringify(data)
            res = self.app.conn.execute(self.app.sessions_table.update().where(self.app.sessions_table.c.session_id == session.sid)
                .values(data=data, last_modification_date=datetime.datetime.now()))
            if res.rowcount == 0:
                self.app.conn.execute(self.app.sessions_table.insert().values(session_id=session.sid, data=data))

class SessionHandler(object):
    def __init__(self, app):
        self.app = app
        self.session_store = DbSessionStore(app)
        self._stack = werkzeug.local.LocalStack()
        self.current = self._stack()

    @contextlib.contextmanager
    def declare_session(self, sid=None):
        if sid is None:
            session = self.session_store.new()
        else:
            session = self.session_store.get(sid)
        self._stack.push(session)
        try:
            yield session
        finally:
            if session.should_save:
                self.session_store.save(session)
            self._stack.pop()



# Copyright (c) 2014, Nicolas Vanhoren
# 
# Released under the MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals, print_function, absolute_import

import asgard
import mailflash
import os.path
import jinja2

class MailsPlugin(asgard.Plugin):

    config_key = "mails"
    dependencies = []

    def __init__(self, app):
        self.app = app
        self.mail = mailflash.Mail()
        self.template_folder = "email_templates"
        self._jinja_env = None

    def configure(self, config):
        if "template_folder" in config:
            self.template_folder = config["template_folder"]
            del config["template_folder"]
        self.mail.init_from_dict(config)

    @property
    def jinja_env(self):
        if self._jinja_env is None:
            self._jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(
                os.path.join(self.app.root_path, self.template_folder)))
        return self._jinja_env

    def build_message(self,
                template,
                template_params=None,
                subject=None,
                recipients=None,
                sender=None,
                cc=None,
                bcc=None,
                attachments=None,
                reply_to=None,
                date=None,
                charset=None,
                extra_headers=None,
                mail_options=None,
                rcpt_options=None):
        """Creates a message by rendering a template.

        :param template: Name of the jinja2 template.
        :param template_params: Parameters of the template (dictionary).
        :param subject: email subject header
        :param recipients: list of email addresses
        :param body: plain text message
        :param html: HTML message
        :param sender: email sender address, or **MAIL_DEFAULT_SENDER** by default
        :param cc: CC list
        :param bcc: BCC list
        :param attachments: list of Attachment instances
        :param reply_to: reply-to address
        :param date: send date
        :param charset: message character set
        :param extra_headers: A dictionary of additional headers for the message
        :param mail_options: A list of ESMTP options to be used in MAIL FROM command
        :param rcpt_options:  A list of ESMTP options to be used in RCPT commands
        """
        template_params = template_params or {}
        rendered = self.jinja_env.get_template(template).render(**template_params)
        return mailflash.Message(subject, recipients, None, rendered, sender, cc, bcc, attachments,
            reply_to, date, charset, extra_headers, mail_options, rcpt_options)

    def send_message(self, message):
        """Sends a mailflash.Message object."""
        self.mail.send(message)

    def send_mail(self, *args, **kwargs):
        message = self.build_message(*args, **kwargs)
        self.send_message(message)



# Copyright (c) 2014, Nicolas Vanhoren
# 
# Released under the MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals, print_function, absolute_import

import werkzeug.local
import flask.helpers
import sqlalchemy as sa
import contextlib
from . import sessions
from . import web

class Asgard(object):

    def __init__(self, import_name, flask_parameters=None):
        """
        Creates an Asgard application.
        """

        self.config = {}
        self.import_name = import_name
        self._root_path = flask.helpers.get_root_path(import_name)

        """
        The engine used to connect to the database.
        """
        self.engine = None

        self._conn_stack = werkzeug.local.LocalStack()
        """
        A proxy to a connection object currently used to perform calls to the database.
        """
        self.connection = self._conn_stack()
        """
        The metadata object containing all the information about the db schema.
        """
        self.metadata = sa.MetaData()

        self.sessions_table = sessions.create_sessions_table(self.metadata)
        self.session_handler = sessions.SessionHandler(self)
        self.session = werkzeug.local.LocalProxy(lambda: self.session_handler.current)

        flask_parameters = flask_parameters or {}
        self.web_app = web.WebApp(self, self.import_name, **flask_parameters)

        self._plugins = []

    @property
    def root_path(self):
        return self._root_path

    @root_path.setter
    def root_path(self, value):
        self._root_path = value
        self.web_app.root_path = value

    @property
    def conn(self):
        return self.connection

    def configure(self, config):
        config = config or {}
        self.config = config
        if "root_path" in config:
            self.root_path = config["root_path"]
        self.configure_database(self.config.setdefault("database", {}))
        self.configure_web(self.config.setdefault("web", {}))
        for plugin in self._plugins:
            plugin[1].configure(config.setdefault(plugin[0].config_key, {}))

    def configure_database(self, config):
        config.setdefault("sqlalchemy.url", 'sqlite://')
        self.engine = sa.engine_from_config(config)

    def configure_web(self, config):
        self.web_app.config.update(**config)

    def create_tables(self):
        self.metadata.create_all(self.engine)

    @contextlib.contextmanager
    def transaction(self):
        """
        A context manager that initializes a connection and store it in the ``conn`` proxy. When the operations
        terminate normally, the transaction is commited. If there is an exception, the transaction is rollbacked.
        """
        assert self._conn_stack.top is None, "Only one connection can be opened at the same time"
        self._conn_stack.push(self.engine.connect())
        try:
            self.conn.current_transaction = self.conn.begin()
            try:
                yield
                self.conn.current_transaction.commit()
            except:
                self.conn.current_transaction.rollback()
                raise
        finally:
            try:
                self.conn.close()
            except:
                pass
            self._conn_stack.pop()

    def transactional(self, func):
        """
        A decorator that will call ``transaction`` before the invocation of the function.
        """
        def alt(*args, **kwargs):
            with self.transaction():
                return func(*args, **kwargs)
        alt.__name__ = func.__name__
        alt.__module__ = func.__module__
        return alt

    def manager(self, claz):
        """
        A decorator to mark a class as a manager. Thus, a single instance of that class will be instanciated and
        stored in the ``i`` attribute of the class and its methods will be accessible by RPC calls.

        It should be noted that all methods declared in a manager should take in arguments and return only types
        succeptible to be correctly serialized/deserialized by whatever RPC protocol will be used in front of
        the managers.
        """
        instance = claz()
        claz.i = instance
        return claz

    def __enter__(self):
        _app_stack.push(self)
        return self

    def __exit__(self, *args, **kwargs):
        _app_stack.pop()

    def declare_session(self, sid=None):
        return self.session_handler.declare_session(sid)

    def plugin(self, plugin_class):
        found = None
        for p in self._plugins:
            if p[0] == plugin_class:
                return p[1]
        return None

    def register_plugin(self, plugin_class):
        if self.plugin(plugin_class) is not None:
            return self.plugin(plugin_class)
        for dep in plugin_class.dependencies:
            self.register_plugin(dep)
        p = plugin_class(self)
        self._plugins.append((plugin_class, p))
        return p

class Plugin(object):
    
    config_key = None
    dependencies = []

    def __init__(self, app):
        pass

    def configure(self, config):
        pass

_app_stack = werkzeug.local.LocalStack()
"""
A proxy to the current Asgard application.
"""
app = _app_stack()

engine = werkzeug.local.LocalProxy(lambda: app.engine)
conn = werkzeug.local.LocalProxy(lambda: app.connection)
connection = werkzeug.local.LocalProxy(lambda: app.connection)
session = werkzeug.local.LocalProxy(lambda: app.session)
config = werkzeug.local.LocalProxy(lambda: app.config)



# Copyright (c) 2014, Nicolas Vanhoren
# 
# Released under the MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals, print_function, absolute_import

import sjoh.flask
import flask

COOKIE_DURATION = 3 * 7 * 34 * 60 * 60 # 3 weeks

class WebApp(flask.Flask):
    def __init__(self, app, import_name, **kwargs):
        self.app = app
        self.sjoh = sjoh.flask.SjohFlask(self)

        super(WebApp, self).__init__(import_name, **kwargs)

    def full_dispatch_request(self, *args, **kw):
        with self.app:
            sid = flask.request.cookies.get('sid')
            with self.app.declare_session(sid):
                response = super(WebApp, self).full_dispatch_request(*args, **kw)
                response.set_cookie('sid', self.app.session.sid, COOKIE_DURATION)
            return response

    def add_url_rule(self, rule, endpoint=None, view_func=None, *args, **kwargs):
        no_transaction = kwargs.get("no_transaction", False)
        if "no_transaction" in kwargs: del kwargs["no_transaction"]
        trans_func = self.app.transactional(view_func) if not no_transaction else view_func
        trans_func.__name__ = view_func.__name__
        trans_func.__module__ = view_func.__module__

        return super(WebApp, self).add_url_rule(rule, endpoint, trans_func, *args, **kwargs)

    def add_url_rule_for_json(self, rule, endpoint=None, view_func=None, *args, **kwargs):
        no_transaction = kwargs.get("no_transaction", False)
        if "no_transaction" in kwargs: del kwargs["no_transaction"]
        trans_func = self.app.transactional(view_func) if not no_transaction else view_func
        return self.sjoh.add_url_rule_for_json(rule, endpoint, trans_func, *args, no_transaction=True, **kwargs)

    def json(self, rule, **options):
        def decorator(f):
            endpoint = options.pop('endpoint', None)
            self.add_url_rule_for_json(rule, endpoint, f, **options)
            return f
        return decorator




# Copyright (c) 2014, Nicolas Vanhoren
# 
# Released under the MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals, print_function, absolute_import

import os.path
import sqlalchemy as sa
import sqlalchemy.sql as sql
import sqlalchemy.sql.expression
import werkzeug.local
import collections
from . import expression as expr
import re
import datetime
from .application import conn

class PersistenceException(Exception):
    pass

class UnrecoverablePersistenceException(PersistenceException):
    """
        This is a type of exception that occurs when we attempted data modification but not all data
        could be modified correctly. When this exception occurs, it is mandatory to rollback the whole
        transaction or the database could contain inconsistencies.
    """
    pass

_order_regex = re.compile("^([a-zA-Z_][a-zA-Z_0-9]*(?:\.[a-zA-Z_][a-zA-Z_0-9]*)*)(?:\s+(asc|desc))?$")

class TableManager(object):
    table = None

    @property
    def table(self):
        return self.__class__.table

    def _expression(self, expression):
        """
            Converts any type of expression into a valid SqlAlchemy clause element which can be
            inserted into any query using the `where_clause()` method.
        """
        exp, values = _convert_expression(expression)
        if isinstance(exp, sqlalchemy.sql.expression.ClauseElement):
            return exp
        if exp is None:
            return sqlalchemy.sql.expression.literal(True)
        qbh = expr.QueryBuilderHelper(self.table)
        where_clause = qbh.where_clause(exp, values)
        subselect = sql.select([self.table.c.id]).select_from(qbh.from_clause())
        subselect = subselect.where(where_clause)
        return self.table.c.id.in_(subselect)

    def create(self, values=None):
        return self.create_many([values])[0]

    def create_many(self, values_list=None):
        """
        Create multiple records in the table using the given list of dictionaries.
        """
        assert isinstance(values_list, collections.Iterable), "Expected a list: %s" % values_list
        ins = self.table.insert()
        created = []
        for val in values_list:
            val = val or {}
            assert isinstance(val, dict), "Expected a dictionary: %s" % val
            if __debug__:
                for k in val.keys():
                    assert hasattr(self.table.c, k), "Table %s doesn't contain a column named %s" % (self.table, k)
            created.append(conn.execute(ins.values(**val)).inserted_primary_key[0])
        return created

    def read_by_id(self, id, fields=None):
        """
        Returns a dictionary containing the asked fields for the current table where the record is identified by
        the given id.

        If the id is not found in the table this method will raise a ``PersistenceException``.
        """
        assert id is not None, "id can not be None"
        return self.read_many_by_id([id], fields)[0]

    def read_many_by_id(self, ids, fields=None):
        """
        Returns a list of dictionaries containing the asked fields for the current table where records are identified by
        the given ids.

        The records will always be returned in the same order they were asked. If one id is not found in the table
        this method will raise a ``PersistenceException``.
        """
        assert isinstance(ids, collections.Iterable), "Expected a list: %s" % ids
        hasid = fields is None or "id" in fields
        res = self.read(self.table.c.id.in_(ids), fields if hasid else fields + ["id"])
        index = dict([(x["id"], x) for x in res])
        res = []
        for id in ids:
            if id not in index:
                raise PersistenceException("Id %s was not found in table %s" % (id, self.table.name))
            res.append(index[id])
        if not hasid:
            for el in res:
                del el["id"]
        return res

    def read(self, expression=None, fields=None, order=None, limit=None, offset=None):
        if fields is None:
            fields = self.table.c.keys()
        exp, values = _convert_expression(expression)

        qbh = expr.QueryBuilderHelper(self.table)
        # list of fields
        selectable = [qbh.column(k) for k in fields]
        # where clause
        where_clause = qbh.where_clause(exp, values)
        # orders
        order = order or []
        order = order if not isinstance(order, (sqlalchemy.sql.expression.ClauseElement, str, unicode)) else [order]
        order_bys = []
        for o in order:
            if isinstance(o, (str, unicode)):
                match = _order_regex.match(o)
                assert match is not None, "Not a valid order specifier: %s" % o
                o = qbh.column(match.group(1))
                o = o.asc() if match.group(2) != "desc" else o.desc()
            order_bys.append(o)
        # query
        query = sql.select(selectable)
        query = query.select_from(qbh.from_clause())
        if where_clause is not None:
            query = query.where(where_clause)
        for o in order_bys:
            query = query.order_by(o)
        # limit & offeset
        if limit:
            query = query.limit(limit)
        if offset:
            query = query.offset(offset)
        # execution
        res = conn.execute(query)
        # convertion to dict
        lst = []
        for el in res:
            dct = {}
            for i in range(len(fields)):
                dct[fields[i]] = el[selectable[i]]
            lst.append(dct)
        return lst

    def count(self, expression=None):
        exp, values = _convert_expression(expression)
        qbh = expr.QueryBuilderHelper(self.table)
        where_clause = qbh.where_clause(exp, values)
        query = sql.select([sql.func.count(self.table.c.id)])
        if where_clause is not None:
            query = query.where(where_clause)
        res = conn.execute(query).fetchone()[0]
        return res

    def read_and_count(self, expression=None, fields=None, order=None, limit=None, offset=None):
        count = self.count(expression)
        res = self.read(expression, fields, order, limit, offset)
        return [res, count]

    def update_by_id(self, id, values):
        self.update_many_by_id([id], values)

    def update_many_by_id(self, ids, values):
        assert isinstance(ids, collections.Iterable), "Expected a list: %s" % ids
        rowcount = self.update(self.table.c.id.in_(ids), values)
        if rowcount != len(ids):
            raise UnrecoverablePersistenceException("One or more ids where not found while updating table %s", self.table.name)

    def update(self, expression=None, values=None):
        values = values or {}
        assert isinstance(values, dict), "Expected a dictionary: %s" % values
        if __debug__:
            for k in values.keys():
                assert hasattr(self.table.c, k), "Table %s doesn't contain a column named %s" % (self.table, k)
        query = self.table.update()
        query = query.where(self._expression(expression))
        query = query.values(values)
        rowcount = conn.execute(query).rowcount
        return rowcount

    def delete_by_id(self, id):
        self.delete_many_by_id([id])

    def delete_many_by_id(self, ids):
        assert isinstance(ids, collections.Iterable), "Expected a list: %s" % ids
        rowcount = self.delete(self.table.c.id.in_(ids))
        if rowcount != len(ids):
            raise UnrecoverablePersistenceException("One or more ids where not found while deleting rows in table %s", self.table.name)

    def delete(self, expression=None):
        query = self.table.delete()
        query = query.where(self._expression(expression))
        rowcount = conn.execute(query).rowcount
        return rowcount

def _convert_expression(expression):
    if isinstance(expression, sqlalchemy.sql.expression.ClauseElement):
        return expression, {}
    if isinstance(expression, (str, unicode, type(None))):
        return expression, {}
    else:
        assert isinstance(expression, collections.Iterable), "Expected a list: %s" % expression
        assert len(expression) == 2, "Expected a list of 2 elements: %s" % expression
        return expression

def table_manager(table):
    """
    A function creating a class binded to a specific SqlAlchemy table. That class will contain generic methods to ease
    manipulation of that table.
    """
    assert table is not None, "Must provide a table"
    assert hasattr(table.c, "id"), "Table %s must contain a field named id" % table
    assert isinstance(table.c.id.type, sa.Integer), "Field id in table %s must be an integer" % table
    assert table.c.id.primary_key, "Field id in table %s must be a primary key" % table

    ttable = table
    class BindedTableManager(TableManager):
        table = ttable

    return BindedTableManager



# Copyright (c) 2014, Nicolas Vanhoren
# 
# Released under the MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
SAQL
====

A small domain-specific language to create expressions representing SQL where-clauses. Those expressions
will be compiled into SQLAlchemy objects.

The goal is to allow arbitrary clients interrogating a server to specify filters when searching a table.
These expressions are language agnostic (they are just strings) and are sufficiently restrictive to be considered
safe even if they were generated by an unsafe environment.

Documentation of the SAQL Language
----------------------------------

SAQL is *not* SQL. It has characteristics coming from SQL, JavaScript and Python. Example expression:

::
    name like "%Smith%" and account != null and status in ["married", "single"]

Simple Types
~~~~~~~~~~~~

* Integers: ``1``, ``2``, ``3``
* Floats: ``1.2``, ``3.4``, ``5.``
* Strings: ``"some string"``, ``'some other string'``. Both simple and double quotes are considered valid.
* Booleans: ``true``, ``false``
* Null: ``null``

Column Names
~~~~~~~~~~~~

Just specify the name of the column, like ``my_column``.

Variables
~~~~~~~~~~~~

You can specify variables to be inserted by the evaluator. Example:

::
    credit > :min_value

If the evaluator is given the integer ``100`` for ``min_value``, the resulting expression will be ``credit > 100``.
Please note that the evaluator will take care of the escaping of the variables it is given. So here we should
give an integer to it. If we give it the value ``"100"`` (as a string), the resulting expression will be
``credit > "100"`` which will probably result in a invalid expression if those types are not comparable.

Variables are the recommended way to pass values coming from forms or similar stuff. They also are the only way to
specify values for dates and datetimes, as there is no way currently to specify those using the SAQL syntax.

Lists
~~~~~

Lists use the ``[`` and ``]`` delimiters.

::
    [1, 2, 3]

They are mostly used in conjonction with the ``in`` operator.

Operators
~~~~~~~~~

Here is a list of the supported binary operators: or, and, ==, !=, in, like, ilike, <=, >=, <, >, +, -, *, / and %.

Additionaly, ``not`` is supported as an unary operators as well as + and -.

Please note that, contrary to SQL, operators like ``not in`` or ``not like`` are not supported. Instead, you should
use the ``not`` operator in front of the condition to inverse. Example:

::
    not (status in ["married", "single"])

"""

from __future__ import unicode_literals, print_function, absolute_import

import sqlalchemy as sa
import sqlalchemy.sql as sql
import sqlalchemy.sql.expression as expr
import operator
from pyparsing import *
import ast
import pylru

CACHE_SIZE = 200

class _NoneVal(object):
    pass

_none_val = _NoneVal()

def _define_parser():
    ParserElement.enablePackrat()

    identifier_part = Word(alphas + "_", alphanums + "_")
    identifier = delimitedList(identifier_part, ".").setParseAction(lambda x: ('identifier', ".".join(x)))

    variable = Combine(":" + Word(alphas + "_", alphanums + "_")).setParseAction(lambda x: ('variable', x[0][1:]))

    boolean_literal = (Keyword("true") | Keyword("false")).setParseAction(lambda x: x[0] == "true")
    integer_literal = Word(nums).setParseAction(lambda x: int(x[0]))
    float_literal = Combine(Word(nums) + "." + Optional(Word(nums))).setParseAction(lambda x: float(x[0]))
    string_literal = quotedString.setParseAction(lambda x: ast.literal_eval(x[0]))
    null_literal = Keyword("null").setParseAction(lambda x: _none_val)

    literal = (boolean_literal | string_literal | float_literal | integer_literal | null_literal).setParseAction(lambda x: ('literal', x[0]))

    logical_or = Keyword("or")
    logical_and = Keyword("and")
    equality = Literal("==") | Literal("!=") | Keyword("in") | Keyword("like") | Keyword("ilike")
    relational = Literal("<=") | Literal(">=") | Literal("<") | Literal(">")
    additive = Literal("+") | Literal("-")
    mult = Literal("*") | Literal("/") | Literal("%")
    non = Keyword("not")

    expression = Forward()

    # can we put a trailing comma?
    list_expr = ("[" + delimitedList(expression) + "]").setParseAction(lambda x: ('list', x[1:-1]))

    rvalue = literal | identifier | variable | list_expr

    def op1_action(x):
        x = x[0]
        assert len(x) == 2
        return ('op1', x[0], x[1])

    def op2_action(x):
        x = x[0]
        if len(x) == 1:
            return x
        return op2_action([[('op2', x[1], x[0], x[2])] + x[3:]])

    expression << operatorPrecedence(rvalue, [
            (additive, 1, opAssoc.RIGHT, op1_action),
            (non, 1, opAssoc.RIGHT, op1_action),
            (mult, 2, opAssoc.LEFT, op2_action),
            (additive, 2, opAssoc.LEFT, op2_action),
            (relational, 2, opAssoc.LEFT, op2_action),
            (equality, 2, opAssoc.LEFT, op2_action),
            (logical_and, 2, opAssoc.LEFT, op2_action),
            (logical_or, 2, opAssoc.LEFT, op2_action),
        ])

    program = expression + stringEnd

    return program

_parser = _define_parser()

def _in(elem1, elem2):
    assert isinstance(elem1, expr.ColumnElement), "Invalid left operand for 'in' operator: %s" % elem1
    return elem1.in_(elem2)

def _like(elem1, elem2):
    assert isinstance(elem1, expr.ColumnElement), "Invalid left operand for 'like' operator: %s" % elem1
    return elem1.like(elem2)

def _ilike(elem1, elem2):
    assert isinstance(elem1, expr.ColumnElement), "Invalid left operand for 'ilike' operator: %s" % elem1
    return elem1.ilike(elem2)

_operators = {
    "or": expr.or_,
    "and": expr.and_,
    "==": operator.eq,
    "!=": operator.ne,
    "in": _in,
    "<=": operator.le,
    ">=": operator.ge,
    "<": operator.lt,
    ">": operator.gt,
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "/": operator.div,
    "%": operator.mod,
    "like": _like,
    "ilike": _ilike,
}

_cache = pylru.lrucache(CACHE_SIZE)

def _compile(expression):
    try:
        return _cache[expression]
    except KeyError:
        pass
    tree = _parser.parseString(expression)[0]
    _cache[expression] = tree
    return tree

class QueryBuilderHelper(object):
    def __init__(self, table):
        """
        :param name: The SqlAlchemy Metadata object.
        :param cache_size: The size of the cache to store compilation results. Can be deactivated by using ``0``.
        """
        assert hasattr(table.c, "id"), "Table %s must contain a column named id" % table
        self.table = table
        self.fk_columns = {}

    def where_clause(self, expression, values=None):
        values = values or {}
        if isinstance(expression, expr.ClauseElement):
            return expression
        elif expression is None:
            return None
        tree = _compile(expression)
        where_clause = self._walk(values, tree)
        if not isinstance(where_clause, expr.ClauseElement):
            where_clause = expr.literal(where_clause)
        return where_clause

    def _walk(self, values, elem):
        kind = elem[0]
        val = elem[1:]
        if kind == "identifier":
            return self.column(val[0])
        elif kind == "variable":
            return values[val[0]]
        elif kind == "literal":
            return val[0] if val[0] != _none_val else None
        elif kind == "list":
            nlist = []
            for el in val[0]:
                nlist.append(self._walk(values, el))
            return nlist
        elif kind == "op1":
            return self._op1(values, *val)
        elif kind == "op2":
            return self._op2(values, *val)
        assert False, "should not happen"

    def column(self, val):
        vals = val.split(".")
        return self._column_walk(self, vals)

    def _column_walk(self, current, vals):
        table = current.table
        column_name = vals[0]
        assert hasattr(table.c, column_name), "Table %s doesn't contain a column named %s" % (table, column_name)
        if len(vals) == 1:
            return getattr(table.c, column_name)
        fk_columns = current.fk_columns
        if column_name not in fk_columns:
            col = getattr(table.c, column_name)
            assert len(col.foreign_keys) > 0, "Column %s is not a foreign key" % col
            assert len(col.foreign_keys) < 2, "Column %s has multiple foreign keys attached to it, this is not supported" % col
            other_column = list(col.foreign_keys)[0].column
            other_table = other_column.table
            assert other_column is other_table.c.id, "Column %s in table %s must reference the column id in table %s" % (col, table, other_table)
            table_alias = other_table.alias()
            fk_columns[column_name] = _JoinPart()
            fk_columns[column_name].table = table_alias
            fk_columns[column_name].fk_columns = {}
        # an assertion to inform about an easy-to-avoid bug, it could be a good idea to fix this one day or later
        assert vals[1] != "id", "Querying the id of a row through a foreign key is not supported, use the foreign key instead"
        return self._column_walk(fk_columns[column_name], vals[1:])

    def from_clause(self):
        return self._walk_tables(self.table, self)

    def _walk_tables(self, current_from, ctx):
        foreign_keys = sorted(ctx.fk_columns.keys())
        for fk in foreign_keys:
            fkctx = ctx.fk_columns[fk]
            current_from = current_from.outerjoin(fkctx.table, fkctx.table.c.id == getattr(ctx.table.c, fk))
            current_from = self._walk_tables(current_from, fkctx)
        return current_from

    def _op1(self, values, op, elem):
        elem = self._walk(values, elem)
        if op == "+":
            return + elem
        elif op == "-":
            return - elem
        elif op == "not":
            return expr.not_(elem)
        assert False, "should not happen"

    def _op2(self, values, op, elem1, elem2):
        elem1 = self._walk(values, elem1)
        elem2 = self._walk(values, elem2)
        assert op in _operators.keys(), "Unsupported operator: %s" % op
        return _operators[op](elem1, elem2)

class _JoinPart(object):
    pass
